"""Release [0.73.0].

Revision ID: 0.73.0
Revises: 0.72.0
Create Date: 2025-01-21 15:49:51.336822

"""

# revision identifiers, used by Alembic.
revision = "0.73.0"
down_revision = "0.72.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
