"""Release [0.80.0].

Revision ID: 0.80.0
Revises: 12eff0206201
Create Date: 2025-03-19 10:07:36.959628

"""

# revision identifiers, used by Alembic.
revision = "0.80.0"
down_revision = "12eff0206201"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
