function r(n,t){if(!n||!t)return"Not available";const e=new Date(n),o=new Date(t);if(isNaN(e.getTime())||isNaN(o.getTime()))return"";const s=Math.abs(o.getTime()-e.getTime()),i=Math.floor(s/(1e3*60)),a=Math.floor(s%(1e3*60)/1e3);return`${i}min ${a}s`}function c(n){const t=new Date,e=new Date;return e.setMonth(t.getMonth()-6),n<e}function l(n){const t=new Date,e=new Date;return e.setFullYear(t.getFullYear()-1),n<e}function u(n){const t=Math.floor(n/60),e=Math.floor(n%60);return`${t}min ${e}s`}export{l as a,r as c,c as i,u as s};
