"""Release [0.57.0].

Revision ID: 0.57.0
Revises: 0.57.0rc2
Create Date: 2024-05-02 13:05:56.733082

"""

# revision identifiers, used by Alembic.
revision = "0.57.0"
down_revision = "0.57.0rc2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
