import{j as e}from"./@radix-B1sy0Lhr.js";import{S as a}from"./chevron-right-double-CbMG1dOM.js";import{c as n,bp as i,B as o}from"./index-CO6UN3UX.js";function h({className:s,children:r,...t}){return e.jsxs("div",{className:n("flex h-9 items-center border-b border-theme-border-moderate bg-theme-surface-primary px-4 py-3",s),...t,children:[e.jsx(i,{asChild:!0,children:e.jsxs(o,{type:"button",className:"flex aspect-square items-center justify-center p-0",intent:"secondary",emphasis:"minimal",children:[e.jsx(a,{className:"h-5 w-5 fill-neutral-500"}),e.jsx("span",{className:"sr-only",children:"Close"})]})}),r]})}export{h as S};
