"""Release [0.84.0].

Revision ID: 0.84.0
Revises: 360fa84718bf
Create Date: 2025-07-10 14:07:30.176139

"""

# revision identifiers, used by Alembic.
revision = "0.84.0"
down_revision = "360fa84718bf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
