"""Release [0.74.0].

Revision ID: 0.74.0
Revises: 4d5524b92a30
Create Date: 2025-02-06 08:10:51.286515

"""

# revision identifiers, used by Alembic.
revision = "0.74.0"
down_revision = "4d5524b92a30"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
