"""Release [0.56.0].

Revision ID: 0.56.0
Revises: 1a9a9d2a836d
Create Date: 2024-03-20 13:30:40.013587

"""

# revision identifiers, used by Alembic.
revision = "0.56.0"
down_revision = "1a9a9d2a836d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
