import{j as s}from"./@radix-B1sy0Lhr.js";import{R as e,U as n,V as l,W as t}from"./index-CO6UN3UX.js";function h({connectorType:o}){const r=o.logo_url,i=o.name;return s.jsx(e,{children:s.jsxs(n,{children:[s.jsx(l,{children:s.jsxs("div",{className:"size-5 shrink-0",children:[s.jsx("img",{width:24,height:24,className:"size-5 shrink-0",src:r??"",alt:`Logo of ${i}`}),s.jsx("p",{className:"sr-only",children:i})]})}),s.jsx(t,{children:i})]})})}function d({resourceType:o}){const r=o.logo_url,i=o.name;return s.jsx(e,{children:s.jsxs(n,{children:[s.jsx(l,{children:s.jsxs("div",{className:"size-5 shrink-0",children:[s.jsx("img",{width:24,height:24,className:"size-5 shrink-0",src:r??"",alt:`Logo of ${i}`}),s.jsx("p",{className:"sr-only",children:i})]})}),s.jsx(t,{children:i})]})})}export{h as C,d as R};
