"""Release [0.80.2].

Revision ID: 0.80.2
Revises: 6611d4bcc95b
Create Date: 2025-04-08 16:18:18.112107

"""

# revision identifiers, used by Alembic.
revision = "0.80.2"
down_revision = "6611d4bcc95b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
