"""Remove user hub token [909550c7c4da].

Revision ID: 909550c7c4da
Revises: 0.63.0
Create Date: 2024-08-05 16:02:48.990897

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "909550c7c4da"
down_revision = "0.63.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.drop_column("hub_token")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("hub_token", sa.VARCHAR(), nullable=True)
        )

    # ### end Alembic commands ###
