import{j as r}from"./@radix-B1sy0Lhr.js";import{R as o}from"./RunsBody-CZAiSxYK.js";import{b as t}from"./ButtonGroup-DVhzbQxV.js";import"./refresh-C8BXS3Fb.js";import"./Pagination-C1sE5Nzn.js";import"./index-CO6UN3UX.js";import"./@tanstack-Dwlisomv.js";import"./@react-router-CHjLNlgw.js";import"./@reactflow-gbyyU3kq.js";import"./chevron-right-double-CbMG1dOM.js";import"./SearchField-Byv1PtST.js";import"./index-g7wOyc86.js";import"./all-pipeline-runs-query-CdvWtiAY.js";import"./terminal-C2tAF2XG.js";import"./DisplayDate-57lUNS79.js";import"./ExecutionStatus-CQbWovhV.js";import"./check-circle-CG6XAdk-.js";import"./help-CRPfbPTO.js";import"./trash-D9LA4VFD.js";import"./AlertDialogDropdownItem-BpLj419i.js";import"./DeleteAlertDialog-CzhTVQaC.js";import"./zod-D48zuELD.js";import"./index.esm-SBF9nn_P.js";import"./bulk-delete-wdObjfps.js";import"./delete-run-XSre8ru-.js";function k(){return r.jsx(t,{children:r.jsx(o,{})})}export{k as default};
