"""Release [0.35.1].

Revision ID: 0.35.1
Revises: 0.35.0
Create Date: 2023-03-07 15:15:42.057028

"""

# revision identifiers, used by Alembic.
revision = "0.35.1"
down_revision = "0.35.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
