import{j as s}from"./@radix-B1sy0Lhr.js";import{bo as c,c as i}from"./index-CO6UN3UX.js";import{S as e}from"./check-C9Nye5lR.js";function n({tickClasses:r,...o}){return s.jsx(c,{...o,children:s.jsx(e,{className:i("h-4 w-4 fill-success-300",r)})})}export{n as T};
