"""Release [0.84.1].

Revision ID: 0.84.1
Revises: 8d4b9ba22c1f
Create Date: 2025-07-29 08:56:44.756559

"""

# revision identifiers, used by Alembic.
revision = "0.84.1"
down_revision = "8d4b9ba22c1f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
