"""Release [0.70.0].

Revision ID: 0.70.0
Revises: 904464ea4041
Create Date: 2024-11-12 12:20:05.537646

"""

# revision identifiers, used by Alembic.
revision = "0.70.0"
down_revision = "904464ea4041"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
