import{j as o}from"./@radix-B1sy0Lhr.js";import{C as r,S as m}from"./StackComponentList-D85oab98.js";import{a as i}from"./@react-router-CHjLNlgw.js";import"./plus-DOrkJCmy.js";import"./refresh-C8BXS3Fb.js";import"./Pagination-C1sE5Nzn.js";import"./index-CO6UN3UX.js";import"./@tanstack-Dwlisomv.js";import"./@reactflow-gbyyU3kq.js";import"./chevron-right-double-CbMG1dOM.js";import"./SearchField-Byv1PtST.js";import"./index-g7wOyc86.js";import"./DisplayDate-57lUNS79.js";import"./expand-full-C8p_0jQv.js";import"./SheetHeader-Bb3v9rha.js";import"./Tabs-2uwqXsdL.js";import"./terminal-C2tAF2XG.js";import"./chevron-down-Bl7WwQAL.js";import"./transform-S7omcvTm.js";import"./NestedCollapsible-0yviIfaA.js";import"./eye-BXcQevds.js";import"./CollapsibleCard-DSZzbjx5.js";import"./check-circle-CG6XAdk-.js";import"./CodeSnippet-B7Wd4R8u.js";import"./all-pipeline-runs-query-CdvWtiAY.js";import"./ExecutionStatus-CQbWovhV.js";import"./help-CRPfbPTO.js";import"./index-BYHxFvoG.js";import"./stack-detail-query-DnyMCdVE.js";import"./ComponentBadge-MVT08kBV.js";import"./ComponentIcon-C9AY8usJ.js";import"./bar-chart-square-check-_m-oYysN.js";import"./layout-GIiNvS10.js";import"./trash-D9LA4VFD.js";import"./DeleteAlertDialog-CzhTVQaC.js";import"./zod-D48zuELD.js";import"./index.esm-SBF9nn_P.js";import"./Infobox-DnduZjNU.js";import"./info-CI2szPUG.js";import"./bulk-delete-wdObjfps.js";function O(){const{connectorId:t}=i();return o.jsx(r,{children:o.jsx(m,{fixedQueryParams:{connector_id:t},displayCreateComponent:!1})})}export{O as default};
