import{j as o}from"./@radix-B1sy0Lhr.js";import{S as r}from"./Tabs-2uwqXsdL.js";import{R as m}from"./RunsBody-CZAiSxYK.js";import{b as i}from"./ButtonGroup-DVhzbQxV.js";import{S as p}from"./StackList-DSAjbVb5.js";import{a as s}from"./@react-router-CHjLNlgw.js";import"./index-CO6UN3UX.js";import"./@tanstack-Dwlisomv.js";import"./@reactflow-gbyyU3kq.js";import"./terminal-C2tAF2XG.js";import"./chevron-down-Bl7WwQAL.js";import"./transform-S7omcvTm.js";import"./DisplayDate-57lUNS79.js";import"./NestedCollapsible-0yviIfaA.js";import"./eye-BXcQevds.js";import"./CollapsibleCard-DSZzbjx5.js";import"./check-circle-CG6XAdk-.js";import"./CodeSnippet-B7Wd4R8u.js";import"./refresh-C8BXS3Fb.js";import"./Pagination-C1sE5Nzn.js";import"./chevron-right-double-CbMG1dOM.js";import"./SearchField-Byv1PtST.js";import"./index-g7wOyc86.js";import"./all-pipeline-runs-query-CdvWtiAY.js";import"./ExecutionStatus-CQbWovhV.js";import"./help-CRPfbPTO.js";import"./trash-D9LA4VFD.js";import"./AlertDialogDropdownItem-BpLj419i.js";import"./DeleteAlertDialog-CzhTVQaC.js";import"./zod-D48zuELD.js";import"./index.esm-SBF9nn_P.js";import"./bulk-delete-wdObjfps.js";import"./delete-run-XSre8ru-.js";import"./plus-DOrkJCmy.js";import"./index-BYHxFvoG.js";import"./stack-detail-query-DnyMCdVE.js";import"./SheetHeader-Bb3v9rha.js";import"./components-Br2ezRib.js";import"./NumberBox-C2Pju4PR.js";import"./ComponentBadge-MVT08kBV.js";import"./ComponentIcon-C9AY8usJ.js";import"./bar-chart-square-check-_m-oYysN.js";import"./layout-GIiNvS10.js";import"./Infobox-DnduZjNU.js";import"./info-CI2szPUG.js";import"./DialogItem-DD_7v3UY.js";function $(){const{componentId:t}=s();return o.jsx("div",{className:"@container",children:o.jsx(r,{isPanel:!1,stacksTabContent:o.jsx(p,{fixedQueryParams:{component_id:t}}),runsTabContent:o.jsx(i,{children:o.jsx(m,{fixedQueryParams:{stack_component:t}})}),componentId:t})})}export{$ as default};
