import{j as s}from"./@radix-B1sy0Lhr.js";import{T as i,c as o}from"./index-CO6UN3UX.js";function c({resourceTypes:m,className:t,size:r="sm",iconSize:a=20}){return s.jsx("div",{className:"flex flex-wrap gap-1",children:m.map(e=>s.jsxs(i,{size:r,rounded:!1,emphasis:"minimal",className:o("w-fit gap-0.5 text-theme-text-primary",t),color:"grey",children:[s.jsx("img",{width:a,height:a,className:"shrink-0 object-contain",src:e.logo_url??"",alt:`Logo of ${e.name}`}),s.jsx("p",{children:e.name})]},e.name))})}export{c as R};
