"""Release [0.46.1].

Revision ID: 0.46.1
Revises: 93cbda80a732
Create Date: 2023-11-10 15:45:26.180743

"""

# revision identifiers, used by Alembic.
revision = "0.46.1"
down_revision = "93cbda80a732"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
