import{j as e,r}from"./@radix-B1sy0Lhr.js";import{P as o,a as n,b as l,c as i,d as m,e as c}from"./ProCta-CUkpV3PJ.js";import{k as d,bx as u}from"./index-CO6UN3UX.js";import{C as x}from"./CodeSnippet-B7Wd4R8u.js";import{f as p}from"./@react-router-CHjLNlgw.js";import"./check-C9Nye5lR.js";import"./@tanstack-Dwlisomv.js";import"./@reactflow-gbyyU3kq.js";const f="/assets/mcp-Cb1aMeoq.webp";function h(){const[a]=p(),t=a.get("model");function s(){return t?`zenml model list --name='contains:${t}'`:"zenml model list"}return e.jsxs(d,{className:"flex flex-wrap justify-between p-5",children:[e.jsxs("div",{children:[e.jsx("p",{className:"text-text-xl font-semibold",children:"Staying Open Source? "}),e.jsx("p",{className:"text-theme-text-secondary",children:"No problem! Use this CLI command to list your models"})]}),e.jsx(x,{code:s()})]})}function v(){const{setTourState:a,tourState:{tourActive:t}}=u();return r.useEffect(()=>{t&&a(s=>({...s,run:!0,stepIndex:s.stepIndex}))},[t]),e.jsx("div",{children:e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsxs(o,{className:"relative overflow-y-hidden",children:[e.jsxs("div",{className:"w-full max-w-none space-y-5 lg:max-w-[900px]",children:[e.jsx(n,{children:"Access Advanced Model Management Features with ZenML Pro"}),e.jsx(l,{}),e.jsx(i,{features:[{title:"Model Control Plane Dashboard",subtitle:"Centralized model management and monitoring"},{title:"Enterprise Security",subtitle:"Social SSO, RBAC, and User Management"},{title:"Managed ZenML Server",subtitle:"On your VPC or hosted on our infrastructure"},{title:"Advanced MLOps",subtitle:"CI/CD/CT, Artifact Control Plane and more"}]}),e.jsx(m,{})]}),e.jsx(c,{className:"absolute translate-x-[30%] translate-y-[15%] scale-[40%]",src:f,alt:"Screenshot of the ZenML Pro Artifact Control Plane"})]}),e.jsx(h,{})]})})}export{v as default};
