"""Release [0.55.4].

Revision ID: 0.55.4
Revises: 479103df60b6
Create Date: 2024-02-29 14:03:34.556396

"""

# revision identifiers, used by Alembic.
revision = "0.55.4"
down_revision = "479103df60b6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
