"""Release [0.64.0].

Revision ID: 0.64.0
Revises: 3dcc5d20e82f
Create Date: 2024-08-08 12:25:12.058636

"""

# revision identifiers, used by Alembic.
revision = "0.64.0"
down_revision = "3dcc5d20e82f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
