"""Release [0.71.0].

Revision ID: 0.71.0
Revises: cc269488e5a9
Create Date: 2024-12-04 20:44:19.316139

"""

# revision identifiers, used by Alembic.
revision = "0.71.0"
down_revision = "cc269488e5a9"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
