"""Release [0.54.1].

Revision ID: 0.54.1
Revises: 7b651bf6822e
Create Date: 2024-01-15 15:10:54.004327

"""

# revision identifiers, used by Alembic.
revision = "0.54.1"
down_revision = "7b651bf6822e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
