"""Release [0.44.3].

Revision ID: 0.44.3
Revises: 0.44.2
Create Date: 2023-09-26 08:35:32.889649

"""

# revision identifiers, used by Alembic.
revision = "0.44.3"
down_revision = "0.44.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
