import{j as e}from"./@radix-B1sy0Lhr.js";import{k as t,B as a,r as n}from"./index-CO6UN3UX.js";import{L as i}from"./@react-router-CHjLNlgw.js";import"./@tanstack-Dwlisomv.js";import"./@reactflow-gbyyU3kq.js";const l="/assets/templates-1S_8WeSK.webp",o=["Run pipelines from the dashboard with Templates","Re-run a pipeline easily from the UI","Model and Artifact Control Plane Dashboard","Managed ZenML server on your VPC or hosted on our servers","Social SSO, RBAC, and User Management","CI/CD/CT, and more!"];function m(){return e.jsxs(t,{className:"relative overflow-hidden border border-warning-100 px-7 py-5",children:[e.jsxs("div",{className:"max-w-[450px] space-y-4",children:[e.jsx("h2",{className:"text-display-xs font-semibold",children:"Access Advanced Template Features with ZenML Pro"}),e.jsx("ul",{className:"space-y-2",children:o.map((r,s)=>e.jsx("li",{className:"text-text-md text-theme-text-secondary",children:r},s))}),e.jsxs("div",{className:"flex flex-wrap items-center gap-2",children:[e.jsx(a,{asChild:!0,className:"truncate bg-warning-600 hover:bg-warning-400 active:bg-warning-500 active:ring-warning-50",size:"md",children:e.jsx(i,{to:n.upgrade,children:"Try Pro Features"})}),e.jsx(a,{emphasis:"minimal",size:"md",children:e.jsx("a",{href:"https://www.zenml.io/pro",target:"_blank",rel:"noopener noreferrer",children:"Learn more"})})]})]}),e.jsx("img",{className:"absolute right-0 top-0 hidden translate-x-[30%] scale-75 md:translate-y-[30%] lg:block xl:-translate-y-[5%]",src:l,alt:"Screenshot of Zenml Pro Templates Feature"})]})}function u(){return e.jsx("div",{className:"layout-container space-y-5 pt-5",children:e.jsx(m,{})})}export{u as default};
