"""Release [0.53.0].

Revision ID: 0.53.0
Revises: 1041bc644e0d
Create Date: 2023-12-20 10:19:02.618928

"""

# revision identifiers, used by Alembic.
revision = "0.53.0"
down_revision = "1041bc644e0d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
