"""Release [0.62.0].

Revision ID: 0.62.0
Revises: b4fca5241eea
Create Date: 2024-07-15 14:15:45.347033

"""

# revision identifiers, used by Alembic.
revision = "0.62.0"
down_revision = "b4fca5241eea"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
