import{x as s,F as r,G as i,H as u}from"./index-CO6UN3UX.js";import{c}from"./@tanstack-Dwlisomv.js";async function h(a){const e=r(u.login),t=await s(e,{method:"POST",headers:{"Content-Type":"application/x-www-form-urlencoded"},body:new URLSearchParams(a)});if(!t.ok){const o=await t.json().then(n=>n.detail).catch(()=>["","Failed to login"]);throw new i({status:t.status,statusText:t.statusText,message:o[1]||"Failed to login"})}return t.json()}function m(a){return c({mutationFn:async e=>h(e),...a})}export{m as u};
