import{F as n,H as r,Z as a,x as u,_ as o,G as i}from"./index-CO6UN3UX.js";import{b as c}from"./@tanstack-Dwlisomv.js";function l({params:e}){return["runs",e]}async function p({params:e}){const s=n(r.runs.all+"?"+a(e)),t=await u(s,{method:"GET",headers:{"Content-Type":"application/json"}});if(t.status===404&&o(),!t.ok)throw new i({message:"Error while fetching pipeline runs",status:t.status,statusText:t.statusText});return t.json()}function y(e,s){return c({queryKey:l(e),queryFn:()=>p(e),...s})}export{y as u};
