"""Release [0.68.0].

Revision ID: 0.68.0
Revises: 1d8f30c54477
Create Date: 2024-10-24 13:14:56.912159

"""

# revision identifiers, used by Alembic.
revision = "0.68.0"
down_revision = "1d8f30c54477"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
