"""Remove component spec path [b73bc71f1106].

Revision ID: b73bc71f1106
Revises: ec6307720f92
Create Date: 2024-11-29 09:36:33.089945

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b73bc71f1106"
down_revision = "ec6307720f92"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.drop_column("component_spec_path")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("component_spec_path", sa.VARCHAR(), nullable=True)
        )

    # ### end Alembic commands ###
