import{j as r}from"./@radix-B1sy0Lhr.js";import{F as s}from"./Wizard-2FIi8JNY.js";import{a}from"./ProviderRadio-C6pLVNxC.js";function n({children:o,displayCancel:t=!0}){return r.jsxs(s,{children:[t&&r.jsx(a,{}),o]})}export{n as S};
