import{j as e}from"./@radix-B1sy0Lhr.js";import{w as r,a as s}from"./index-CO6UN3UX.js";import{E as l}from"./EmptyState-BMA34iVR.js";function c({err:t,isAlertCircle:a=!1}){return e.jsx(l,{icon:a?e.jsx(r,{className:"h-[120px] w-[120px] fill-neutral-300"}):e.jsx(s,{className:"h-[120px] w-[120px] fill-neutral-300"}),children:e.jsx("div",{className:"text-center",children:e.jsx("p",{className:"max-w-full whitespace-pre-wrap break-words text-text-lg text-theme-text-secondary",children:t.message})})})}export{c as E};
