import{r as m,j as e}from"./@radix-B1sy0Lhr.js";import{aH as p,aI as d,aw as g}from"./index-CO6UN3UX.js";const h=m.forwardRef((r,a)=>{const{triggerChildren:n,children:i,onSelect:o,onOpenChange:t,icon:x,open:s,...l}=r;return e.jsxs(p,{open:s,onOpenChange:t,children:[e.jsx(d,{asChild:!0,children:e.jsx(g,{...l,className:"hover:cursor-pointer",icon:r.icon,ref:a,onSelect:c=>{c.preventDefault(),o&&o()},children:n})}),i]})});h.displayName="DialogItem";export{h as D};
