"""Release [0.65.0].

Revision ID: 0.65.0
Revises: bf2120261b5a
Create Date: 2024-08-28 15:42:13.858104

"""

# revision identifiers, used by Alembic.
revision = "0.65.0"
down_revision = "bf2120261b5a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
