import{x as n,F as o,G as i,H as u}from"./index-CO6UN3UX.js";import{c}from"./@tanstack-Dwlisomv.js";async function p(r){const e=o(u.currentUser),t=await n(e,{method:"PUT",headers:{"Content-Type":"application/json"},body:JSON.stringify(r)});if(!t.ok){const s=await t.json().then(a=>Array.isArray(a.detail)?a.detail[1]:a.detail).catch(()=>"Failed to update User");throw new i({status:t.status,statusText:t.statusText,message:s})}return t.json()}function f(r){return c({mutationFn:async e=>p(e),...r})}export{f as u};
